@@ -6,24 +6,33 @@ Django internals are instrumented via normal `patch()`.
 `django.apps.registry.Apps.populate` is patched to add instrumentation for any
 specific Django apps like Django Rest Framework (DRF).
 """
+from inspect import getmro
+from inspect import isclass
+from inspect import isfunction
 import sys
 
-from inspect import isclass, isfunction, getmro
-
-from ddtrace import config, Pin
-from ddtrace.vendor import debtcollector, six, wrapt
+from ddtrace import Pin
+from ddtrace import config
 from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
-from ddtrace.contrib import func_name, dbapi
-from ddtrace.ext import http, sql as sqlx, SpanTypes
-from ddtrace.http import store_request_headers, store_response_headers
+from ddtrace.constants import SPAN_MEASURED_KEY
+from ddtrace.contrib import dbapi
+from ddtrace.contrib import func_name
+from ddtrace.ext import SpanTypes
+from ddtrace.ext import http
+from ddtrace.ext import sql as sqlx
 from ddtrace.internal.logger import get_logger
-from ddtrace.propagation.http import HTTPPropagator
 from ddtrace.propagation.utils import from_wsgi_header
-from ddtrace.utils.formats import asbool, get_env
+from ddtrace.utils.formats import asbool
+from ddtrace.utils.formats import get_env
+from ddtrace.vendor import debtcollector
+from ddtrace.vendor import six
+from ddtrace.vendor import wrapt
 
+from . import conf
+from . import utils
 from .. import trace_utils
-from .compat import get_resolver, user_is_authenticated
-from . import utils, conf
+from .compat import get_resolver
+from .compat import user_is_authenticated
 
 
 log = get_logger(__name__)
@@ -43,12 +52,11 @@ config._add(
         analytics_sample_rate=None,
         trace_query_string=None,  # Default to global config
         include_user_name=True,
-        use_handler_resource_format=get_env("django", "use_handler_resource_format", default=False),
+        use_handler_resource_format=asbool(get_env("django", "use_handler_resource_format", default=False)),
+        use_legacy_resource_format=asbool(get_env("django", "use_legacy_resource_format", default=False)),
     ),
 )
 
-propagator = HTTPPropagator()
-
 
 def patch_conn(django, conn):
     def cursor(django, pin, func, instance, args, kwargs):
@@ -74,50 +82,43 @@ def patch_conn(django, conn):
 
 
 def instrument_dbs(django):
-    def all_connections(wrapped, instance, args, kwargs):
-        conns = wrapped(*args, **kwargs)
-        for conn in conns:
-            try:
-                patch_conn(django, conn)
-            except Exception:
-                log.debug("Error instrumenting database connection %r", conn, exc_info=True)
-        return conns
+    def set_connection(wrapped, instance, args, kwargs):
+        _, conn = args
+        try:
+            patch_conn(django, conn)
+        except Exception:
+            log.debug("Error instrumenting database connection %r", conn, exc_info=True)
+        return wrapped(*args, **kwargs)
 
-    if not isinstance(django.db.connections.all, wrapt.ObjectProxy):
-        django.db.connections.all = wrapt.FunctionWrapper(django.db.connections.all, all_connections)
+    if not isinstance(django.db.connections.__setitem__, wrapt.ObjectProxy):
+        django.db.connections.__setitem__ = wrapt.FunctionWrapper(django.db.connections.__setitem__, set_connection)
 
     if hasattr(django.db, "connection") and not isinstance(django.db.connection.cursor, wrapt.ObjectProxy):
         patch_conn(django, django.db.connection)
 
 
 def _set_request_tags(django, span, request):
-    span.set_tag("django.request.class", func_name(request))
-    span.set_tag(http.METHOD, request.method)
-
-    if django.VERSION >= (2, 2, 0):
-        headers = request.headers
-    else:
-        headers = {}
-        for header, value in request.META.items():
-            name = from_wsgi_header(header)
-            if name:
-                headers[name] = value
-
-    store_request_headers(headers, span, config.django)
+    span._set_str_tag("django.request.class", func_name(request))
 
     user = getattr(request, "user", None)
     if user is not None:
-        if hasattr(user, "is_authenticated"):
-            span.set_tag("django.user.is_authenticated", user_is_authenticated(user))
+        # Note: getattr calls to user / user_is_authenticated may result in ImproperlyConfigured exceptions from
+        # Django's get_user_model():
+        # https://github.com/django/django/blob/a464ead29db8bf6a27a5291cad9eb3f0f3f0472b/django/contrib/auth/__init__.py
+        try:
+            if hasattr(user, "is_authenticated"):
+                span._set_str_tag("django.user.is_authenticated", str(user_is_authenticated(user)))
 
-        uid = getattr(user, "pk", None)
-        if uid:
-            span.set_tag("django.user.id", uid)
+            uid = getattr(user, "pk", None)
+            if uid:
+                span._set_str_tag("django.user.id", str(uid))
 
-        if config.django.include_user_name:
-            username = getattr(user, "username", None)
-            if username:
-                span.set_tag("django.user.name", username)
+            if config.django.include_user_name:
+                username = getattr(user, "username", None)
+                if username:
+                    span._set_str_tag("django.user.name", username)
+        except Exception:
+            log.debug("Error retrieving authentication information for user %r", user, exc_info=True)
 
 
 @trace_utils.with_traced_module
@@ -130,11 +131,11 @@ def traced_cache(django, pin, func, instance, args, kwargs):
         # update the resource name and tag the cache backend
         span.resource = utils.resource_from_cache_prefix(func_name(func), instance)
         cache_backend = "{}.{}".format(instance.__module__, instance.__class__.__name__)
-        span.set_tag("django.cache.backend", cache_backend)
+        span._set_str_tag("django.cache.backend", cache_backend)
 
         if args:
             keys = utils.quantize_key_values(args[0])
-            span.set_tag("django.cache.key", keys)
+            span._set_str_tag("django.cache.key", str(keys))
 
         return func(*args, **kwargs)
 
@@ -215,11 +216,14 @@ def traced_populate(django, pin, func, instance, args, kwargs):
     return ret
 
 
-def traced_func(django, name, resource=None):
+def traced_func(django, name, resource=None, ignored_excs=None):
     """Returns a function to trace Django functions."""
 
     def wrapped(django, pin, func, instance, args, kwargs):
-        with pin.tracer.trace(name, resource=resource):
+        with pin.tracer.trace(name, resource=resource) as s:
+            if ignored_excs:
+                for exc in ignored_excs:
+                    s._ignore_exception(exc)
             return func(*args, **kwargs)
 
     return trace_utils.with_traced_module(wrapped)(django)
@@ -268,7 +272,11 @@ def traced_load_middleware(django, pin, func, instance, args, kwargs):
             def wrapped_factory(func, instance, args, kwargs):
                 # r is the middleware handler function returned from the factory
                 r = func(*args, **kwargs)
-                return wrapt.FunctionWrapper(r, traced_func(django, "django.middleware", resource=mw_path))
+                if r:
+                    return wrapt.FunctionWrapper(r, traced_func(django, "django.middleware", resource=mw_path))
+                # If r is an empty middleware function (i.e. returns None), don't wrap since NoneType cannot be called
+                else:
+                    return r
 
             trace_utils.wrap(base, attr, wrapped_factory)
 
@@ -312,10 +320,7 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
     try:
         request_headers = request.META
 
-        if config.django.distributed_tracing_enabled:
-            context = propagator.extract(request_headers)
-            if context.trace_id:
-                pin.tracer.context_provider.activate(context)
+        trace_utils.activate_distributed_headers(pin.tracer, int_config=config.django, request_headers=request_headers)
 
         # Determine the resolver and resource name for this request
         resolver = get_resolver(getattr(request, "urlconf", None))
@@ -338,6 +343,8 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
 
             if config.django.use_handler_resource_format:
                 resource_format = "{method} {handler}"
+            elif config.django.use_legacy_resource_format:
+                resource_format = "{handler}"
             else:
                 # In Django >= 2.2.0 we can access the original route or regex pattern
                 # TODO: Validate if `resolver.pattern.regex.pattern` is available on django<2.2
@@ -371,18 +378,16 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
             "django.request",
             resource=resource,
             service=trace_utils.int_service(pin, config.django),
-            span_type=SpanTypes.HTTP,
+            span_type=SpanTypes.WEB,
         ) as span:
+            span.metrics[SPAN_MEASURED_KEY] = 1
             analytics_sr = config.django.get_analytics_sample_rate(use_global_config=True)
             if analytics_sr is not None:
                 span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, analytics_sr)
 
-            if config.django.http.trace_query_string:
-                span.set_tag(http.QUERY_STRING, request_headers["QUERY_STRING"])
-
             # Not a 404 request
             if resolver_match:
-                span.set_tag("django.view", resolver_match.view_name)
+                span._set_str_tag("django.view", resolver_match.view_name)
                 utils.set_tag_array(span, "django.namespace", resolver_match.namespaces)
 
                 # Django >= 2.0.0
@@ -390,11 +395,9 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
                     utils.set_tag_array(span, "django.app", resolver_match.app_names)
 
             if route:
-                span.set_tag("http.route", route)
+                span._set_str_tag("http.route", route)
 
             # Set HTTP Request tags
-            span.set_tag(http.URL, utils.get_request_uri(request))
-
             response = func(*args, **kwargs)
 
             # Note: this call must be done after the function call because
@@ -403,10 +406,8 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
             _set_request_tags(django, span, request)
 
             if response:
-                span.set_tag(http.STATUS_CODE, response.status_code)
-                if 500 <= response.status_code < 600:
-                    span.error = 1
-                span.set_tag("django.response.class", func_name(response))
+                status = response.status_code
+                span._set_str_tag("django.response.class", func_name(response))
                 if hasattr(response, "template_name"):
                     # template_name is a bit of a misnomer, as it could be any of:
                     # a list of strings, a tuple of strings, a single string, or an instance of Template
@@ -434,8 +435,28 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
 
                     utils.set_tag_array(span, "django.response.template", template_names)
 
-                headers = dict(response.items())
-                store_response_headers(headers, span, config.django)
+                url = utils.get_request_uri(request)
+
+                if django.VERSION >= (2, 2, 0):
+                    request_headers = request.headers
+                else:
+                    request_headers = {}
+                    for header, value in request.META.items():
+                        name = from_wsgi_header(header)
+                        if name:
+                            request_headers[name] = value
+
+                response_headers = dict(response.items())
+                trace_utils.set_http_meta(
+                    span,
+                    config.django,
+                    method=request.method,
+                    url=url,
+                    status_code=status,
+                    query=request.META.get("QUERY_STRING", None),
+                    request_headers=request_headers,
+                    response_headers=response_headers,
+                )
 
             return response
 
@@ -451,10 +472,10 @@ def traced_template_render(django, pin, wrapped, instance, args, kwargs):
 
     with pin.tracer.trace("django.template.render", resource=resource, span_type=http.TEMPLATE) as span:
         if template_name:
-            span.set_tag("django.template.name", template_name)
+            span._set_str_tag("django.template.name", template_name)
         engine = getattr(instance, "engine", None)
         if engine:
-            span.set_tag("django.template.engine.class", func_name(engine))
+            span._set_str_tag("django.template.engine.class", func_name(engine))
 
         return wrapped(*args, **kwargs)
 
@@ -512,7 +533,9 @@ def _instrument_view(django, view):
 
     # If the view itself is not wrapped, wrap it
     if not isinstance(view, wrapt.ObjectProxy):
-        view = wrapt.FunctionWrapper(view, traced_func(django, "django.view", resource=func_name(view)))
+        view = wrapt.FunctionWrapper(
+            view, traced_func(django, "django.view", resource=func_name(view), ignored_excs=[django.http.Http404])
+        )
     return view
 
 
@@ -589,8 +612,8 @@ def patch():
 
 def _unpatch(django):
     trace_utils.unwrap(django.apps.registry.Apps, "populate")
-    trace_utils.unwrap(django.core.handlers.base.BaseHandler, "lotrace_utils.ad_middleware")
-    trace_utils.unwrap(django.core.handlers.base.BaseHandler, "getrace_utils.t_response")
+    trace_utils.unwrap(django.core.handlers.base.BaseHandler, "load_middleware")
+    trace_utils.unwrap(django.core.handlers.base.BaseHandler, "get_response")
     trace_utils.unwrap(django.template.base.Template, "render")
     trace_utils.unwrap(django.conf.urls.static, "static")
     trace_utils.unwrap(django.conf.urls, "url")
